/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Agora.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
          (C) 2003  Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef AGORA_H
#define AGORA_H

#include <arpa/inet.h>

#ifdef WIN32
#include <ws2tcpip.h>
#endif

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <DCDT_Defs.h>

#include <DCDT_Msg.h>

// TODO effort to check if possible to use signalling with C++
#include <signal.h>

class DCDT_Member;
class DCDT_PostOffice;
class DCDT_Finder;
class DCDT_MsgManager;
class DCDT_InnerLinkManager; 

/*! \brief Structure that contains all the informations about an Agora member
 *
 *
 */
typedef struct _DCDT_MemberTableElem {

  /// Flag: false -> User member; true -> System member
  bool system;

  /// Flag: true -> Free; false -> Used
  bool empty;

  /// Flag: false -> Inactive; true -> A thread is running this member
  bool active;

  /// Mutex used when accessing this data structure
  DCDT_Mutex mtx;

  /// Flag: true -> Free; false -> Used
  bool FreeSlot;

  DCDT_Member * MemberPtr;
  unsigned int Profile;
  int status;
  int num_missed_deadlines; // Not actually used
} DCDT_MemberTableElem;

typedef struct _FinderData {
  /// Unicast interface address ( string )
  char IfStrAddr[INET_ADDRSTRLEN];
  /// Unicast interface address in network mode
  int IfNumAddr;
  /// Multicast interface address ( string )
  char McStrAddr[INET_ADDRSTRLEN];
  /// Multicast interface address in network mode
  int McNumAddr;
  /// Port for multicast and unicast address written in network mode
  short McPort;
  /// File descriptor for the configuration file
  FILE *fp;
} FinderData;

// in order to use signal handler
typedef void (*sighandler_t) (int);


/*! \brief Main dcdt object, virtualize and manage threads.
 *
 *  Each project must have at least one DCDT_Agora object.\n
 *  System Members started automatically:\n
 *  - DCDT_Finder: manage communication channels with external Agora.\n
 *  - DCDT_InnerLinkManager: talks with others Agora running on the local machine.\n
 *  - DCDT_MsgManager: handles message queue.\n
 */

class DCDT_Agora {

 public:
  DCDT_Agora( int PO_Type = 1 );
  DCDT_Agora( char *filename, int PO_TYPE = 1 );
  ~DCDT_Agora();

  void AddMember(DCDT_Member *);
  void AddSysMember( DCDT_Member *, unsigned int profile);
  void RemoveMember(DCDT_Member *);

  int ActivateMember(DCDT_Member *, char *name=NULL );
  int ActivateSysMember(DCDT_Member *, char *name=NULL );
  void LetsWork();

  void SetStatus(int NewStatus);

  inline bool CheckMemberProfile(int MemberID, unsigned int mask);
  inline int GetStatus();
  int ReadAgoraID() { return AgoraID; };
  void ReadIPStrAddr(char * addr) { strcpy(addr, IPStrAddr); };
  int ReadIPAddr() { return IPAddr; };

  ///Check if all instantiated Members are initialized and ready.
  inline int CheckIfAllMemberReady(); 

  inline int GetFirstFreeMemberTableSlot();

  inline void DeleteMemberTableEntry(int MemberID);

  /// Flag: = 1 -> First agora running on this machine
  bool FirstAgora;

  /// Flag: false -> STANDALONE mode, true -> NETWORK mode
  bool commflag;

  inline void SetActive( int MembID)  { MemberTable[MembID].active = true; }; 
  inline void SetInactive( int MembID ) {  MemberTable[MembID].active = false; };

 private:
  
  bool CommConfig(char *filename);
  char IPStrAddr[INET_ADDRSTRLEN];

  /// Agora unique identification based on the IP address, this is valid only for an Agora that is running a Finder ( i.e. the first Agora running in network mode )
  int AgoraID;
  int IPAddr;
  int Status;

  /** Max ID of the created Members */
  int MaxIDMember;

  int PostOfficeType;
  DCDT_PostOffice * PostOffice;
  DCDT_MsgManager * MsgManager;
  DCDT_Finder * Finder;
  
  // TODO: Check if we must keep always the following
  DCDT_InnerLinkManager *InnerLinkManager; 

  int LastDCDT_MemberID;

  DCDT_MemberTableElem MemberTable[MAX_MEMBERS];

  FinderData *finder_data;

#ifdef SEQUENCER_VERSION
  DCDT_Sequencer * Sequencer;
#endif
};

int DCDT_Agora::GetStatus() {
return Status;
}

/*! \brief Check if the profile for the member \a MemberID is \a mask
 *
 *  \param MemberID Member identifier of the member to check
 *  \param mask profile to check
 *  \return true - the member has the correct profile
 *  \return false - the member hasn't the correct profile 
 */ 
inline bool DCDT_Agora::CheckMemberProfile(int MemberID, unsigned int mask) {
  return ((MemberTable[MemberID].Profile & mask) == mask);
}

/*! \brief Return the first free slot or 0 if there is no room.
 *
 */
inline int  DCDT_Agora::GetFirstFreeMemberTableSlot(){
  for( int i = 1; i < MAX_MEMBERS; i++) {
    MemberTable[i].mtx.lock();
    if ( MemberTable[i].FreeSlot) {
      MemberTable[i].FreeSlot = false;
      MemberTable[i].mtx.unlock();
      return i;
      }
    MemberTable[i].mtx.unlock();
    }
  return 0;
};

/*! \brief Frees the slot for the member \a MemberID
 *
 */
inline void  DCDT_Agora::DeleteMemberTableEntry(int MemberID){
  MemberTable[MemberID].mtx.lock();

  MemberTable[MemberID].MemberPtr = NULL;
  MemberTable[MemberID].num_missed_deadlines = 0;
  MemberTable[MemberID].FreeSlot = true;
  MemberTable[MemberID].empty = true;

  MemberTable[MemberID].mtx.unlock();

};


#endif //define
