/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Finder.cpp

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
          (C) 2006   Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef FINDER_CPP
#define FINDER_CPP

using namespace std;

#include <string.h> // for strerror

#include <DCDT_Finder.h>
#include <DCDT_Agora.h>

DCDT_Finder::DCDT_Finder(DCDT_Agora* agora, FinderData *fd) : DCDT_Member(agora)
{
  for(int i=1; i<MAXLINKS; i++) {
    DeleteLinkTableSlot(i);
  }

  for (int i=0; i<MAXCHANNELS; i++) {
    dynamic_channels[i].active = false;
    dynamic_channels[i].channel = NULL;
  }

  hs_header = NULL;
  localCD = remoteCD = NULL;
  finder_data = fd;
  starting_timer = false;
  passed=GetTime()-GetTime();
}


DCDT_Finder::~DCDT_Finder()
{
  if ( finder_data ) {
    if ( finder_data->fp )
      fclose ( finder_data->fp );

    delete finder_data;
  }

  for (int idx=0; idx<MAXCHANNELS; idx++ ) {
    if ( dynamic_channels[idx].channel ) {
      delete dynamic_channels[idx].channel;
    }
  }
}


/*! \brief Initialize dynamic and static channels data structures.
 *
 *  The dynamic channel ( channel 0 ) is the multicast channel used to talk with others Agora connected to the local network.\n
 *  If there are static links delay the initialization of the dynamic channel of STARTING_TIMER_VAL microseconds; the last action is to send an HS_NOTIFY message throught all the active channels.
 */
void DCDT_Finder::Init()
{
  // Open the channel used for dynamic communications on the multicast address
  try {
    dynamic_channels[0].active = true;
    dynamic_channels[0].dynamic = true;
    dynamic_channels[0].channel = new DCDT_Channel_IP(myAgoraID);
    ((DCDT_Channel_IP*)dynamic_channels[0].channel)->DynamicPrepare(finder_data->IfNumAddr, finder_data->McNumAddr, finder_data->McPort);
  } catch (ChannelError e) {
    cout << "\nError while preparing channel 0 (" << strerror(e.errval) << ")\n";
    exit(1);
  }

  // Create static links data structures
  try {
    StaticLinks();
  } catch (Exception e) {
    cout << "\nError while preparing static links\n";
  }

  if (starting_timer) {
    cout << "\ndeactivating ip channel!!!!!!\n";
    dynamic_channels[0].active = false;
    start_ts = GetTime();
  }

  // Send an HS_NOTIFY message throught the active channels.
  for (i=0; i<MAXCHANNELS; i++) {
    if (dynamic_channels[i].active) {
      try {
        dynamic_channels[i].channel->SendNotify();
      } catch(TimeOut e) {
        cout << "\nCan't send notify on channel " << i << "( Timeout )" << "\n";
      } catch(ChannelError e) {
        cout << "\nError while sending notify on channel " << i << "(" << strerror(e.errval) << ")\n";
      }
    }
  }
  old_ts = notify_ts = GetTime();
}


/* \brief Manages links and system messages.
 *
 * This member is logically composed of two parts:\n
 * -  a cyclic part that manages HS_NOTIFY and HS_ANSWER messages on all the active channels, if necessary this part can create or close a channel.\n
 * - a periodical part that becomes active every NOTIFY_PERIOD usec and sends HS_NOTIFY throught all the active channels.\n
 */
void DCDT_Finder::DoYourJob(int par)
{
  // spedizione e ricezione di notify e answer sono interne, cos come la gestione delle CommData. Non si tratta pi quindi
  // di DCDT_Msg. La localCD di inizializzazione del canale  ricavata ogni volta con GetStartingCD(), ed  un puntatore
  // per riferimento, in modo che possa risalire dalla preparazione del Link in caso sia modificata.
  // Il timer dei notify  gestito allo stesso modo, le SysReceive sono senza timer e non
  // bloccanti: i canali sono scandagliati uno dopo l'altro e se avanza del tempo il Finder si addormenta per il tempo
  // rimanente
  // localCD is taken from GetStartingCD() so if the an because is a pointer the initial

  ChannelTableElem *channel;
  int idx;

  // Receive and answer to system messages
  linked = received = false;
  
  try {
    for (idx=0; idx<MAXCHANNELS; idx++) {
      channel=&(dynamic_channels[idx]);
      if (channel->active) {
        try {
          channel->channel->ReceiveHS(hs_header, remoteCD);
        } catch (TimeOut t) {
	  // cout << "TimeOut while receiving on channel " << i << "\n";
        } catch (ChannelError e) {
          cout << "Error while receiving on channel " << i << "\n";
        }
	
        if (hs_header) {
          received = true;
          switch (hs_header->type) {
	    
	  case HS_NOTIFY:
            //server
	    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Notify received from AgoraId %i", hs_header->AgoraID));

	    linknum = IsPresent(hs_header->AgoraID);
	    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("IsPresent: %i", linknum));
	    
	    if (!linknum) {
	      channel->channel->ChooseChannel(hs_header->channel);
	      try {
		localCD = CreateLink(hs_header->AgoraID, L_WAITING, channel->channel->GetStartingCD(), NULL, channel->channel->ReadStartingProfile());
		channel->channel->SendAnswer(remoteCD, localCD);
	      } catch ( TimeOut t ) {
		cout << "TimeOut\n";
	      }
	      catch ( ChannelError e ) {
		cout << "ChannelError " << e.errval << "\n";
	      }
	      linked = true;
	    }
	    else {
	      switch (LinkTable[linknum].LinkPtr->ReadStatus()) {
		
	      case L_WAITING:
		channel->channel->SendAnswer(remoteCD, LinkTable[linknum].LinkPtr->GetCommData());
		break;
		
	      case L_LOST:
		channel->channel->ChooseChannel(hs_header->channel);
		localCD = LinkTable[linknum].LinkPtr->Restart(L_WAITING, 
							      channel->channel->GetStartingCD(),
							      NULL);
		channel->channel->SendAnswer(remoteCD, localCD);
		linked = true;
                break;
		
              case L_KILL:
                DeleteLinkTableSlot( linknum );
                myAgora->RemoveMember(LinkTable[linknum].LinkPtr);
                break;
	      }
	    }
	    delete remoteCD;
	    remoteCD = NULL;
	    break;
	    
	  case HS_ANSWER:
            //client
	    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Answer received from AgoraId %i", hs_header->AgoraID));

	    linknum = IsPresent(hs_header->AgoraID);
	    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("IsPresent: %i", linknum));

	    if (!linknum) {
	      channel->channel->ChooseChannel(hs_header->channel);
	      CreateLink(hs_header->AgoraID, L_CONNECTING, channel->channel->GetStartingCD(), remoteCD, channel->channel->ReadStartingProfile());
	      remoteCD = NULL;
	      linked = true;
	    }
	    else {
              TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Link: %i Status: %d", linknum, LinkTable[linknum].LinkPtr->ReadStatus() ));
	      switch (LinkTable[linknum].LinkPtr->ReadStatus()) {
		
	      case L_WAITING:
		TRC_PRINT( DCDT_TRC_COMM, TRC1, ("RemoteAgoraID=%i MyAgoraID=%i", hs_header->AgoraID, myAgoraID));

		if (hs_header->AgoraID > myAgoraID) {
		  LinkTable[linknum].LinkPtr->ResetToConnecting(channel->channel->GetStartingCD(), remoteCD);
		  remoteCD = NULL;
		}
		else {
		  delete remoteCD;
		  remoteCD = NULL;
		}
		break;
		
	      case L_CONNECTING:
	      case L_RESETTOCONN:
		delete remoteCD;
		remoteCD = NULL;
		break;
		
	      case L_LOST:
		channel->channel->ChooseChannel(hs_header->channel);
		LinkTable[linknum].LinkPtr->Restart(L_CONNECTING, 
						    channel->channel->GetStartingCD(), 
						    remoteCD);
		remoteCD = NULL;
		linked = true;
	      }
	    }
	    break;
	    
	  default:
	    cout << "\nFinder received unknown msg!";
          }
          delete hs_header;
          hs_header = NULL;
        }
      }
    }

    new_ts = GetTime();
    
    // multicast channel reactivation
    if ((starting_timer) && ((new_ts - start_ts) > STARTING_TIMER_VAL)) {
      dynamic_channels[0].active = true;
      starting_timer = false;
      cout << "\nreactivating ip channel!!!!!!\n";
    }

    // Periodical part:
    // If we haven't received a message and our period wasn't expired -> wait
    // Advise others agora that we are alive
    if (!received || ((passed = new_ts - old_ts) >= FINDER_PERIOD)) {
      if ((passed) < FINDER_PERIOD)
        Delay(FINDER_PERIOD - passed);
      
      old_ts = GetTime();
      
      // If the previous NS_NOTIFY message was sent more than NOTIFY_PERIOD usec ago start sendinga new message throught all the active channels.
      if ((new_ts - notify_ts) >= NOTIFY_PERIOD) {
        for (i=0; i<MAXCHANNELS; i++) {
          if (dynamic_channels[i].active) {
            try {
              dynamic_channels[i].channel->SendNotify();
            } catch(TimeOut e) {
              cout << "Can't send notify on channel " << i << "\n";
            } catch(ChannelError e) {
              cout << "Error while sending notify on channel " << i << "\n";
            }
          }
        }
        notify_ts = old_ts;
      }
    }
  } catch (Exception e) {
    cout << "Finder exception!\n";
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Finder exception"));
    //exit(0);
  }
}


/*! \brief Create a new LinkRx system member and active it.
 */
CommData* DCDT_Finder::CreateLink(int r_id, int status, CommData *l_cd, CommData *r_cd, unsigned int profile)
{
  DCDT_LinkRx *newLink;
  int newLinkID;
  
  newLinkID = GetFirstFreeLinkTableSlot();
  newLink = new DCDT_LinkRx(myAgora, newLinkID, r_id);
  myAgora->AddSysMember(newLink, profile);
  LinkTable[newLinkID].LinkPtr = newLink;
  try {
    l_cd = newLink->Prepare(status, l_cd, r_cd);
  } catch ( ChannelError e ) {
    cout << "ChannelError on prepare " <<e.errval << "\n";
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Link created!"));
  myAgora->ActivateMember(newLink);
  return l_cd;
}

/*! \brief Check if there is link to the 'id' agora.
 */
int DCDT_Finder::IsPresent(int id)
{
  for (int i=1; i<MAXLINKS; i++) {
    if ((LinkTable[i].freeSlot == false) && ((LinkTable[i].LinkPtr->ReadRemoteID()) == id))
      return i;
  }
  return 0;
}


void DCDT_Finder::CloseLink(int id) {
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Closing LINK %i", id));

  /*
  if (id < 1000) {
    DeleteLinkTableSlot(id);
    myAgora->RemoveMember(LinkTable[id].LinkPtr);
  }
  */
}

/*! \brief Initialize data structures for static channels and open the communications channels needed.\n
 * 
 *  For IP channel open a socket and bind it to the address written in the configuration file.\n
 * Set starting_timer to true if at least a static channel was successfully opened.\n
 * \exception Exception Error reding the configuration file
 */
void DCDT_Finder::StaticLinks() {
  char ch_type[4], lk_type[15];
  int ch_count = 1, rv;
  char l_ip_addr[INET_ADDRSTRLEN], r_ip_addr[INET_ADDRSTRLEN];
  int l_ip_port, r_ip_port;
  unsigned int profile;
  char device_name[UNIX_PATH_MAX];

  /* 
   * <CHANNEL TYPE> <"link" or "gateway" or "bridge"> <LEVEL> <LOCAL IP> <LOCAL PORT> <REMOTE IP> <REMOTE PORT>\n
   * <CHANNEL TYPE> can be be: IP, SER ...\n
   * "link" = ( ??? )\n
   * "bridge" = ( ??? )\n
   * Ex:TCP link 1 131.175.127.61 1234 198.177.56.9 1444\n
   * NOTE: If the configuration file contains invalid settings the system starts in STANDALONE mode\n
   */
  
  rv = fscanf(finder_data->fp, "%s%s", ch_type, lk_type);
  if (((rv > 0) && (rv < 2))) {
    fclose(finder_data->fp);
    throw Exception();
  }
  while (!feof(finder_data->fp)) {
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("%s %s", ch_type, lk_type));

    profile = MPM_SYSTEM;
    if (!strcmp(lk_type, "link"))
      profile |= MPM_LINK;
    if (!strcmp(lk_type, "bridge"))
      profile |= MPM_BRIDGE;
    if (!strcmp(ch_type, "ip")) {
      rv = fscanf(finder_data->fp, "%s%i%s%i\n", l_ip_addr, &l_ip_port, r_ip_addr, &r_ip_port);
      if ((rv < 4) || (rv == EOF)) {
        fclose(finder_data->fp);
        throw Exception();
      }

      TRC_PRINT( DCDT_TRC_COMM, TRC1, (" %s %i %s %i", l_ip_addr, l_ip_port, r_ip_addr, r_ip_port));

      dynamic_channels[ch_count].active = true;
      dynamic_channels[ch_count].dynamic = false;
      dynamic_channels[ch_count].channel = new DCDT_Channel_IP(myAgoraID);
      ((DCDT_Channel_IP*)dynamic_channels[ch_count].channel)->StaticPrepare(profile, l_ip_addr, htons((short)l_ip_port), r_ip_addr, htons((short)r_ip_port));

      ch_count++;
    }
    else if (!strcmp(ch_type, "ser")) {
      rv = fscanf(finder_data->fp, "%s", device_name);
      if ((rv == 0) || (rv == EOF)) {
        fclose(finder_data->fp);
        throw Exception();
      }

      TRC_PRINT( DCDT_TRC_COMM, TRC1, (" %s", device_name));
      dynamic_channels[ch_count].active = true;
      dynamic_channels[ch_count].dynamic = false;
      dynamic_channels[ch_count].channel = new DCDT_Channel_Serial(myAgoraID);
      ((DCDT_Channel_Serial*)dynamic_channels[ch_count].channel)->StaticPrepare(profile, device_name);

      ch_count++;
    }

    rv = fscanf(finder_data->fp, "%s%s", ch_type, lk_type);
    if ((rv > 0) && (rv < 2)) {
      fclose(finder_data->fp);
      throw Exception();
    }
  }

  if (--ch_count)
    starting_timer = true;
}

#endif //define
