/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Socket.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef SOCKET_H
#define SOCKET_H

#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/socket.h>
#include <errno.h>
#include <stdlib.h>
#include <DCDT_Defs.h>
#include <DCDT_Msg.h>
#include <DCDT_ctrace.h>

/*! \brief Virtual class for socket operations.
 *
 *  A socket is always created in blocking mode.\n
 *  Ports are stored as short like in the socket API, not as int such as in CommData, out of the library the numeric values are always expressed in network format.\n
 *  Ex: 192.168.26.250:1500 -> 0xf01aa8c0:0xdc05\n
 */  
class Socket
{
public:
  Socket()
  {
    sockfd = opened = bound = connected = tconn_flag = tlost_flag = tsend_flag = trecv_flag = 0;
    tconn_value.tv_sec = tconn_value.tv_usec = tlost_value.tv_sec = tlost_value.tv_usec = 0;
    tsend_value.tv_sec = tsend_value.tv_usec = trecv_value.tv_sec = trecv_value.tv_usec = 0;
    unblock_send_flag = unblock_recv_flag = 0;
  };
  virtual ~Socket() {};

  virtual void Open(int fd = 0) = 0;
  inline void Bind(int addr, short port);
  inline void Connect(int addr, short port);
  virtual void Send(const DCDT_Msg *msg) = 0;
  virtual DCDT_Msg* Receive() = 0;
  inline void Close();
  inline void ForcedClose();
  inline void GetOption(int level, int optname);
  inline void SetOption(int level, int optname);
  inline void UnSetOption(int level, int optname);
  inline void GetName(char *addr, short &port);
  inline void GetCommData(int &addr, short &port);
  inline void SetLostTimer(int usec);
  inline void LostTimerOn();
  inline void LostTimerOff();
  inline void SetConnTimer(int usec);
  inline void ConnTimerOn();
  inline void ConnTimerOff();
  inline void SetSendTimer(int usec);
  inline void SendTimerOn();
  inline void SendTimerOff();
  inline void UnblockSend();
  inline void SetReceiveTimer(int usec);
  inline void ReceiveTimerOn();
  inline void ReceiveTimerOff();
  inline void UnblockReceive();
  int GetFd() { return sockfd; };

protected:
  int sockfd,
    /// = 1 -> socket opened
    opened,
    /// = 1 -> socket bound
    bound, 
    /// = 1 -> socket connected
    connected,
    /// = 1 -> use a timer in accept()
    tconn_flag,
    /// = 1 -> use a timer in communications
    tlost_flag, 
    ///= 1 -> use a timer when sending messages
    tsend_flag,
    /// = 1 -> use a timer when receiving messages
    trecv_flag;
  int
    /// = 1 -> socket non blocking when sending
    unblock_send_flag,
    /// = 1 -> socket non blocking when receiving
    unblock_recv_flag;
  struct timeval tconn_value, tlost_value, tsend_value, trecv_value;
  int my_addr;
  short my_port;

};

/*! \brief Bind the socket to the specified address and port.
 *
 *  If port=0 the port to bind to is randomly choose by the kernel, the same for address. Can throw a ChannelError exception if the bind can't be done; on success set bound to 1.
 *  \exception ChannelError
 */
inline void Socket::Bind(int addr, short port)
{
  struct sockaddr_in bindaddr;
  if (!opened || bound) {
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Socket already bound %d or not opened %d", !opened, bound ));
    return;
  }

  bindaddr.sin_family = AF_INET;
  if (!addr)
    bindaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  else {
    bindaddr.sin_addr.s_addr = (in_addr_t)addr;
    my_addr = (int)bindaddr.sin_addr.s_addr;
  }
  bindaddr.sin_port = my_port = port;

  if (bind(sockfd, (struct sockaddr*) &bindaddr, sizeof(bindaddr)) < 0) {
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Error binding socket %d (addr %x:%x)",errno,addr,port ));
    throw ChannelError(errno);
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Socket successfully bound to: %x port: %x", addr, port));

  bound = 1;
}

/*! \brief TCP: connect to the specified address and port, UDP: accept messages only from the specified source.
 *
 *  If the socket is a tcp one connect to the specified address and port, if it's an udp one receive messages only from the specified address and port. Can throw ConnError exception in case of problems. When the connection is successfully established set connected to 1.
 *  \exception ConnError
 */
inline void Socket::Connect(int addr, short port)
{
  struct sockaddr_in connaddr;

  if (!opened || connected)
    return;

  connaddr.sin_family = AF_INET;
  connaddr.sin_addr.s_addr = (in_addr_t)addr;
  connaddr.sin_port = port;

  if (connect(sockfd, (struct sockaddr *) &connaddr, sizeof(connaddr)) < 0)
    throw ConnError(errno);

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Socket %s successfully connected to: %x port: %x", (sockfd==SOCK_DGRAM)?"UDP":"TCP", addr, port));
  connected = 1;
}

/*! \exception ConnError */
inline void Socket::Close()
{
  if (opened) {
    if (close(sockfd) < 0)
      throw ConnError(errno);
    connected = opened = bound = sockfd = 0;

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Socket successfully closed"));
  }
}

inline void Socket::ForcedClose()
{
  if (opened) {
    shutdown(sockfd, SHUT_RDWR);
    connected = opened = bound = sockfd = 0;
    
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Socket successfully shut"));
  }
}


/*! \exception ChannelError */
inline void Socket::GetOption(int level, int optname)
{
  int val;
  socklen_t sl = sizeof(val);

  if (!opened)
    return;
  if (getsockopt(sockfd, level, optname, &val, &sl) < 0)
    throw ChannelError(errno);

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Socket option value: %i", val));
}

/*! \exception ChannelError */
inline void Socket::SetOption(int level, int optname)
{
  int val = 1;

  if (!opened)
    return;
  if (setsockopt(sockfd, level, optname, &val, sizeof(val)) < 0)
    throw ChannelError(errno);

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Socket option successfully set"));
}

/*! \exception ChannelError */
inline void Socket::UnSetOption(int level, int optname)
{
  int val = 0;

  if (!opened)
    throw ChannelError();
  if (setsockopt(sockfd, level, optname, &val, sizeof(val)) < 0)
    throw ChannelError(errno);

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Socket option successfully unset"));
}

/*! \exception ChannelError */
inline void Socket::GetName(char *addr, short &port)
{
  struct sockaddr_in sa;
  socklen_t sl = sizeof(sa);
  char localaddr[INET_ADDRSTRLEN];
  
  if (getsockname(sockfd, (struct sockaddr*) &sa, &sl) < 0 \
      || !inet_ntop(sa.sin_family, &sa.sin_addr, localaddr, INET_ADDRSTRLEN))
    throw ChannelError(errno);
  if (addr)
    strcpy(addr, localaddr);
  port = sa.sin_port;
  
}

/*! \exception ChannelError */
inline void Socket::GetCommData(int &addr, short &port)
{
  struct sockaddr_in sa;
  socklen_t sl = sizeof(sa);

  if (getsockname(sockfd, (struct sockaddr*) &sa, &sl) < 0)
    throw ChannelError(errno);
  if ((int)sa.sin_addr.s_addr == 0)
    addr = my_addr;
  else
    addr = (int)sa.sin_addr.s_addr;
  port = sa.sin_port;

}

inline void Socket::SetLostTimer(int usec)
{
  tlost_value.tv_usec = usec % 1000000;
  tlost_value.tv_sec = usec / 1000000;
};

inline void Socket::LostTimerOn()
{
  tlost_flag = 1;
};

inline void Socket::LostTimerOff()
{
  tlost_flag = 0;
};

inline void Socket::SetConnTimer(int usec)
{
  tconn_value.tv_usec = usec % 1000000;
  tconn_value.tv_sec = usec / 1000000;
};

inline void Socket::ConnTimerOn()
{
  tconn_flag = 1;
};

inline void Socket::ConnTimerOff()
{
  tconn_flag = 0;
};

inline void Socket::SetSendTimer(int usec)
{
  tsend_value.tv_usec = usec % 1000000;
  tsend_value.tv_sec = usec / 1000000;
};

inline void Socket::SendTimerOn()
{
  tsend_flag = 1;
  unblock_send_flag = 0;
};

inline void Socket::SendTimerOff()
{
  tsend_flag = 0;
};

inline void Socket::UnblockSend()
{
  unblock_send_flag = 1;
}

inline void Socket::SetReceiveTimer(int usec)
{
  trecv_value.tv_usec = usec % 1000000;
  trecv_value.tv_sec = usec / 1000000;
};

inline void Socket::ReceiveTimerOn()
{
  trecv_flag = 1;
  unblock_recv_flag = 0;
};

inline void Socket::ReceiveTimerOff()
{
  trecv_flag = 0;
};

inline void Socket::UnblockReceive()
{
  unblock_recv_flag = 1;
}

#endif










