/* EEPROM.C

This file contains all functions to read and write data to pic internal EEPROM.

readByteFromIE: read byte to internal eeprom
readWordFromIE: read word to internal eeprom
writeByteToIE: write byte to internal eeprom
writeWordToIE: write word to internal eeprom
*/

#include <p18f252.h>
#include <stdlib.h>
#include <string.h>
#include "eeprom.h"
#include "definitions.h"

char readByteFromIE(unsigned char address) {
	char dataRead;
	
	EEADR = address; // Data memory address to read
	
	//See MICROCHIP Data Sheet Errata Corrige
	_asm
		nop
	_endasm

	EECON1bits.EEPGD = 0; // point to data memory
	EECON1bits.CFGS = 0; // access program FLASH or Data EEPROM memory
	EECON1bits.RD = 1; // EEPROM read
	
	dataRead = EEDATA; // read data from EEPROM and copy to variable
	
	return dataRead;
}

void writeByteToIE(unsigned char address, char data) {
	EEADR = address; // Data memory address to write
	EEDATA = data; // Data memory value to write
	EECON1bits.EEPGD = 0; // point to data memory
	EECON1bits.CFGS = 0; // access program FLASH 
	EECON1bits.WREN = 1; // enable writes
	
	INTCONbits.GIE = 0; // disable global interrupts
	
	EECON2 = 0x55; 
	EECON2 = 0xAA; // write special data in EECON2 to enable write
	
	EECON1bits.WR = 1; // start write cycle
	
	while (!PIR2bits.EEIF); // wait for write completition
	
	PIR2bits.EEIF = 0; // clear Write Complete Interrupt Flag
	EECON1bits.WREN = 0; // end write
	INTCONbits.GIE = 1; // enable global interrupts
}

//load values from eeprom
unsigned float loadEEPromValue(unsigned char address) {
	unsigned char low_part=0, high_part=0;
	unsigned char exponent=0, sign_exponent=0;
	unsigned char i=0;
	unsigned int temp_value;
	
	unsigned float value = 0.0;
	
	//extract low_part from eeprom
	low_part = (unsigned char)readByteFromIE(address);
	//low_part = 0xff;
	
	//extract high_part from eeprom
	high_part = (unsigned char)readByteFromIE(address+1);
	//high_part = 0xff;
	
	//retrieve value from low_part and high_part
	temp_value = (unsigned int)(high_part&(0xFF>>(SIGN_EXP+EXP)));
	value = (unsigned float)( (temp_value<<8) | (low_part) );
	
	//retrieve exponent
	exponent = (high_part>>(8-SIGN_EXP-EXP))&(0xFF>>(8-SIGN_EXP-EXP+1));
	
	//retrieve sign of the exponent
	sign_exponent = high_part>>7;
	
	//retrieve original value
	for (i=0; i<exponent; i++) {
		if (sign_exponent == NEGATIVE_EXP) 
			value = value/10;
		else
			value = value*10;
	}
	
	//return value;
	return value;
}

//store values in eeprom
void storeEEPromValue(unsigned char address, unsigned float value/*, unsigned char type*/) {
	unsigned int max_value = 1;
	unsigned char i = 0;
	unsigned char low_part=0, high_part=0;
	unsigned char exponent=0, sign_exponent=0;
	unsigned int temp_value = 0;
	
	//calculate max value: it is the greater number we can store directly into the eeprom. It is calculated as 2^VALUE
	max_value = max_value<<VALUE;

	//check if value is integer
	if (value==(int)value) {
		//type is integer
		//if value is greater than max_value, we can store into the eeprom directly
		temp_value = (unsigned int)value;
		
		if (value < max_value) {
			//value is less then max_value
			//set exponent and its sign
			exponent = 0;
			sign_exponent = POSITIVE_EXP;
		}
		else {
			//value is greater then max_value
			//set sign of the exponent
			exponent = 0;
			sign_exponent = POSITIVE_EXP;
			
			for (i=0; i<(1<<EXP); i++) {
				temp_value = (unsigned int)(temp_value/10.0);
				if (temp_value < max_value)
					break;
			}
			//set exponent
			exponent = i;
		}
	}
	else {
		//type is float;
		if (value < max_value) {
			//check if float part is zero
			if (value == (int)value) {
				//the float part is zero: the value is a false float and it van be stored as integer
				//set exponent and its sign
				exponent = 0;
				sign_exponent = POSITIVE_EXP;
			}
			else {
				//the value is really a float because the float part is not null
				
				//set sign of the exponent
				sign_exponent = NEGATIVE_EXP;
				exponent = 0;
		 
				for (i=0; i<((1<<EXP)-1); i++) {
					value = value*10;
					if ((unsigned int)value > max_value) {
						value = value / 10;
						break;
					}
				}
				//set exponent
				exponent = i;
			}
		}
		else {
			//value is greater than max_value
			//the float part will be lost because the integer part is greater than max_value, and we can reduce the integer part to store in the eeprom
			//set sign of the exponent
			exponent = 0;
			sign_exponent = POSITIVE_EXP;
			
			for (i=0; i<(1<<EXP); i++) {
				value = value/10.0;
				if (value < max_value)
					break;
			}
			//set exponent
			exponent = i;
		}
		//set temp_value
		temp_value = (unsigned int)value;
	}

	//create low_part with the 8 less significative bits of the value
	low_part = temp_value&0xFF; 
	//write the low_part in eeprom
	writeByteToIE(address, low_part);
			
	//create high_part with the 5 most significative bits of the value
	high_part = (temp_value>>8)&(0xFF>>(SIGN_EXP+EXP));
	//write the exponent in high_part
	high_part = high_part|(exponent<<(8-SIGN_EXP-EXP));
	//write the sign of exponent in high_part
	high_part = high_part|(sign_exponent<<7);
	//write the high part in eeprom
	writeByteToIE(address+1, high_part);
	
}
