#ifndef ActuationManager_h
#define ActuationManager_h

#ifdef MULTI_AIRBOARD
#include "MultiSerialCommunicationAIRBOARD.h"
#endif

#ifdef EMMEQUADRO
#include "SerialCommunicationEmmeQuadro.h"
#endif

#ifdef ABACO
#include "SerialCommunicationAbaco.h"
#endif

#ifdef AIRBOARD
#include "SerialCommunicationAIRBoard.h"
#endif

#ifdef TRIBOARD
#include "SerialCommunicationTRIBOARD.h"
#endif

/**
 * Type for the lists of data used as interface towards the high-level controller
 */
//using namespace std;

typedef map<string,float> FloatDataListType;

/**
 * This class is the interface for the data sent and received by the high-level
 * controller. 
 *
 * The ActuationManager class is an abstract class (in fact it has two
 * pure virtual methods) that gives the necessary interface to translate
 * the data between the high-level controller and the low-level controller
 * according to the different kinematics (e.g.: holonomous, anholonomous, 
 * synchro, etc.).
 *
 * @short	Interface between high-level and low-level control data
 * @author	Marcello Restelli (restelli@elet.polimi.it)
 */

class ActuationManager 
#ifdef EMMEQUADRO
: public SerialCommunicationEmmeQuadro
#endif
#ifdef MULTI_AIRBOARD
: public MultiSerialCommunicationAIRBoard
#endif
#ifdef AIRBOARD
: public SerialCommunicationAIRBoard
#endif
#ifdef ABACO
: public SerialCommunicationAbaco
#endif
#ifdef TRIBOARD
: public SerialCommunicationTRIBoard
#endif
{
 public:
  // LIFECYCLE

  /**
   * Default constructor
   */
  ActuationManager();

  /**
   * Destructor
   */
  virtual ~ActuationManager();

  // OPERATORS

  //Use compiler-generated copy constructor and assignment
  //ActuationManager(const ActuationManager& from);
  //ActuationManager& operator=(const ActuationManager& from);

  // OPERATIONS

  /**
   * Insert the data from the high-controller into a list
   *
   * @param name	The symbolic name of the datum
   * @param value	The value of the datum
   */
  void InsertData(string name, float value);

  /**
   * Decompose the actions received by the high-level controller
   * into commands for the low-level controller
   */ 
  virtual void ComputeActuations() = 0;

  /**
   * Starting from the data of the encoders, this method computes the robot
   * movement
   */
  virtual void ComputeOdometry() = 0;

 protected:
  /**
   * The list of data received from the high-level controller
   */
  FloatDataListType mCommandData;

  /**
   * The list of data that must be sent to the high-level controller
   */
  FloatDataListType mOdometryData;
};


inline void ActuationManager::InsertData(string name, float value)
{
  mCommandData[name] = value;
}

#endif
