#ifndef SerialCommunicationEmmeQuadro_h
#define SerialCommunicationEmmeQuadro_h

#include "SerialCommunication.h"

#define h2d(h) (h>='0'&&h<='9'?h-'0':10+h-'A')

// Variabili e costanti del modulo
const char hexdec[16] = {'0', '1', '2', '3', '4', '5', '6', '7', 
			 '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};


/**
 * This class implements the communication thorugh serial device toward the
 * control board produced by EmmeQuadro.
 *
 * The SerialCommunicationEmmeQuadro class implements the methods to write to and
 * read from the EmmeQuadro control board. In particular, this class implements
 * two pure virtual methods defined in the @ref LowLevelCommunication interface.
 *
 * @short	Interface for the Emmequadro control board
 * @author	Marcello Restelli (restelli@elet.polimi.it)
 */

class SerialCommunicationEmmeQuadro : public SerialCommunication
{
 public:
  // LIFECYCLE

  /**
   * Default constructor
   */
  SerialCommunicationEmmeQuadro();

  /**
   * Destructor
   */
  virtual ~SerialCommunicationEmmeQuadro();

  // OPERATORS

  //Use compiler-generated copy constructor and assignment
  //SerialCommunicationEmmeQuadro(const SerialCommunicationEmmeQuadro& from);
  //SerialCommunicationEmmeQuadro& operator=(const SerialCommunicationEmmeQuadro& from);

  // OPERATIONS

  
  /**
   * This is a pure virtual method that communicates the data stored in 
   * mActuationValues to the AIRBoard
   */
  void WriteActuations();

  /**
   * This is a pure virtual method that receives the data sent by the AIRBoard
   * and stores them in mEncoderValues
   */
  void ReadEncoders(){};

 protected:

  /**
   * Enable and disable the motors
   *
   * @param i	The motor status (0: disabled, 1: enabled)
   */
  void SetMotorStatus(int i);

  // Reset degli encoders: 0=sx, 1=dx, 2=sx+dx
  /**
   * Reset the counters of the encoders
   *
   * @param i	Which counters must be cleared (0: left, 1:right, 2: both)
   */
  void ResetEncoder(int i);

  /**
   * Read the value of the two encoders
   *
   * @param sx	A pointer that will contain the reading of the left encoder
   * @param dx	A pointer that will contain the reading of the right encoder
   */
  void GetEncoder(long int* sx, long int *dx);

  /**
   * Set the speeds of the two motors
   *
   * @param sx	The speed of the left motor in percentage [-100;100]
   * @param dx	The speed of the right motor in percentage [-100;100]
   */
  void SetMotorsSpeed(int sx, int dx);

  /**
   * Set the status of the two kickers
   *
   * @param i	The status for the front kicker (0: rest, 1: extended)
   * @param j	The status for the rear kicker (0: rest, 1: extended)
   */
  void SetKicker(int i, int j);

  /**
   * Read the status of the bumpers
   *
   * @param dx	Read the status of the right bumper
   * @param sx	Read the status of the left bumper
   */
  void GetBumpers(int *dx, int *sx);

 private:

  // Converte un generico intero "v" nell'equivalente in base2
  /**
   * Maps a value in the range [0;256] to a value in the range [-127;126].
   * In particular, the values in the range [0;126] are left unchanged, while
   * the values in the range [127;256] are mapped into [-127;0]
   *
   * @param v	The value to be converted
   *
   * @result	The converted value 
   */
  int base2 (int v);
  
  /**
   * Maps a value in the range [-127;126] to a value in the range [0;256].
   * In particular, the values in the range [0;126] are left unchanged, while
   * the values in the range [-127;0] are mapped into [127;256]
   *
   * @param v	The value to be converted
   *
   * @result	The converted value 
   */
  unsigned int conv_uint(int i);

  /**
   * Send 4 integer to the control board
   *
   * @param oda0	The first integer to be sent
   * @param oda1	The second integer to be sent
   * @param oda2	The third integer to be sent
   * @param oda3	The fourth integer to be sent
   */
  void Write(int oda0, int oda1, int oda2, int oda3);

  /**
   * Read 4 integer from the control board
   *
   * @param oda0	The first integer that has been read
   * @param oda1	The second integer that has been read
   * @param oda2	The third integer that has been read
   * @param oda3	The fourth integer that has been read
   */
  void Read(int* ida0, int* ida1, int* ida2, int* ida3);
};


inline int SerialCommunicationEmmeQuadro::base2(int v)
{
  if (v>127)
    return (v-256);
  else
    return (v);
}


inline unsigned int SerialCommunicationEmmeQuadro::conv_uint(int i)
{
  if (i<0)
    return (256+i);
  else
    return i;
}


#endif
